use std::io;
use std::os::unix::net::UnixStream;
use std::io::{Read, Write};
use clap::Parser;
use std::num::ParseIntError;
use std::process::{Command, Stdio};
use std::collections::HashSet;

#[derive(Parser, Debug)]
#[command(author, version, about, long_about = None)]
struct Args {
    /// Path to the unix socket for the monitor
    #[arg(long)]
    monitor_socket_path: String,

    /// Path to the unix socket for the uart
    #[arg(long)]
    uart_socket_path: String,

    /// RAM file
    #[arg(long)]
    ram_file: String,
}

fn parse_hex(s: &str) -> Result<u64, ParseIntError> {
    u64::from_str_radix(s.trim_start_matches("0x"), 16)
}

fn print_menu() {
    println!("Menu:");
    println!("1. Show registers");
    println!("2. Update registers");
}

fn read_registers(mut socket: &mut UnixStream) -> Result<String, String> {
    send_qemu_monitor_command(&mut socket, "show_registers");
    let _ = read_until_string(socket, "Register state:");
    let _ = read_until_string(socket, "\n");
    let data = read_until_string(socket, "done");
    let reg_data = &data[..data.len() - 4];
    io::stdout().write(reg_data.as_bytes()).unwrap();
    io::stdout().flush().unwrap();
    Ok(String::from("Success"))
}

fn write_registers(mut socket: &mut UnixStream) -> Result<String, String> {
    println!("List registers:");

    let allowed_register_names = HashSet::from(
        [
            "rax", "rbx", "rcx", "rdx",
            "rbp", "rsp", "rdi", "rsi", "rip",
            "r8", "r9", "r10", "r11", "r12", "r13", "r14", "r15"
        ]);

    let ops : Vec<(String, u64)> = {
        let mut reg_str = String::new();
        io::stdin().read_line(&mut reg_str).expect("Failed to read line");
        let reg_tokens : Vec<&str> = reg_str.trim().split(";").collect();
        let mut update_ops = vec![];
        for token in reg_tokens {
            let set_reg_tokens : Vec<&str> = token.split("=").collect();
            if set_reg_tokens.len() != 2 {
                return Err(format!("Token parsing error {}", token));
            }
            let reg_name = set_reg_tokens[0];
            if !allowed_register_names.contains(reg_name) {
                return Err(format!("Register name '{}' not allowed.", reg_name));
            }
            let reg_value = parse_hex(set_reg_tokens[1]);
            if reg_value.is_err() {
                return Err(format!("Failed to parse value '{}' as u64", set_reg_tokens[1]));
            }
            update_ops.push((String::from(reg_name), reg_value.unwrap()));
        }
        update_ops
    };

    if ops.is_empty() {
        return Err(format!("No ops"));
    }

    let ops_string = ops.iter().map(|p| format!("{}=0x{:x}", p.0, p.1)).collect::<Vec<String>>().join(";");
    let ops_command = format!("update_registers {}", ops_string);

    // Pause the VM
    send_qemu_monitor_command(&mut socket, "stop");

    // Send the update-regs command
    send_qemu_monitor_command(&mut socket, &ops_command);

    // Continue execution
    send_qemu_monitor_command(&mut socket, "nmi");
    send_qemu_monitor_command(&mut socket, "cont");

    Ok(String::from("Success"))
}

fn read_until_string(socket: &mut UnixStream, pattern: &str) -> String {
    let mut data = String::new();
    loop {
        let mut buf = [0u8];
        if let Ok(()) = socket.read_exact(&mut buf) {
            let ch = buf[0] as char;
            data.push(ch);
            if data.ends_with(&pattern[..]) {
                return data;
            }
        }
    }
}

fn send_qemu_monitor_command(socket: &mut UnixStream, command: &str) {
    let _ = read_until_string(socket, "(qemu) ");
    socket.write_all(command.as_bytes()).unwrap();
    socket.write_all(b"\n").unwrap();
}

fn main() {
    let args = Args::parse();

    let ram_size = 256u64 * 1024 * 1024;

    // Launch VM and wait a tiny bit before the sockets are opened
    let mut process = Command::new("./launch_vm.sh")
        .args([
            "--monitor_unix_path", &args.monitor_socket_path,
            "--uart_unix_path", &args.uart_socket_path,
            "--ram_size", &format!("{}", ram_size),
            "--ram_file", &args.ram_file])
        .stdin(Stdio::null())
        .stdout(Stdio::null())
        .stderr(Stdio::null())
        .spawn()
        .expect("Failed to launch VM");
    std::thread::sleep(core::time::Duration::from_secs(1));

    if let Ok(Some(ret)) = process.try_wait() {
        let output = process.wait_with_output().unwrap();
        println!("Process exitted unexpectedly: {}.\nStdout {}.\nStderr {}",
            ret, String::from_utf8(output.stdout).unwrap(), String::from_utf8(output.stderr).unwrap());
        return;
    }

    // Connect to sockets
    let mut monitor_socket = UnixStream::connect(&args.monitor_socket_path).unwrap();
    let mut uart_socket = UnixStream::connect(&args.uart_socket_path).unwrap();

    // Wait for VM to boot
    println!(
        "Awkward, we lost the ssh key to the encrypted VM running on the latest epic HXP hack2 processor.\n\
         The flag is on the VM but how do we get it?\n\n");

    println!("Waiting for the VM to boot... that may take a while ...");
    let _ = read_until_string(&mut uart_socket, "Boot took us");
    read_until_string(&mut uart_socket, "\n");

    println!("Here and ready for hacking!");

    loop {
        print_menu();
        let mut line = String::new();
        io::stdin().read_line(&mut line).expect("Failed to read line");
        match line.trim().parse() {
            Ok(1) => {
                match read_registers(&mut monitor_socket) {
                    Ok(_) => continue,
                    Err(err) => println!("Error: {}", err),
                }
            }
            Ok(2) => {
                match write_registers(&mut monitor_socket) {
                    Ok(_) => continue,
                    Err(err) => println!("Error: {}", err),
                }
            }
            Ok(_) => {
                println!("Unknown input");
                continue;
            }
            Err(_) => {
                println!("Invalid input");
                continue;
            }
        }
        io::stdout().flush().unwrap();
    }
}
